// AutoClient.cpp: Basic S-PLUS command line client 
// Uses Automation to communicate with S-PLUS
//
//////////////////////////////////////////////////////////////////////

#include <afxdisp.h>  // sluggish MFC only needed for COleVariant's
#include <COMDEF.H>	  // _bstr_t
#include <afxpriv.h>  // A2W macros
#include <iostream.h>

#include "AutoClient.h"

#define RESULT_LENGTH	4096

//////////////////////////////////////////////////////////////////////


AutoClient::AutoClient()
{
	
}



AutoClient::~AutoClient()
{
	CoUninitialize();
}



HRESULT AutoClient::CreateSplus()
{
	HRESULT hr;
	
	try
	{
		
		if FAILED(hr = CoInitialize(NULL)) 
			cerr << hex << "Initialization Error: " << hr << endl;
		else
			if FAILED(hr = SPlusApplication.CreateObject(OLESTR("S-PLUS.Application")))
				cerr << hex << "Creation Error: " << hr << endl;
	}
	catch (CException *e)
	{
		cerr << hex << "Invoke Exception: " << e << endl;
		return E_UNEXPECTED;
	}
	
	return hr;
}



HRESULT AutoClient::ExecuteStringResult( LPCTSTR pszCmd, LPTSTR pszResult)
{
	HRESULT hr;
	EXCEPINFO   e;
	UINT     nArgErr;
	
	
	// For the journey ahead we package pszCmd into a variant
	COleVariant vCommand( pszCmd );
	COleVariant vNewStyleFormat( 0L );
	COleVariant vRet;
	VariantInit(&vRet);
	
	try
	{
		hr = SPlusApplication.Invoke(DISPATCH_METHOD, &vRet, &e, &nArgErr,
			OLESTR("ExecuteStringResult"), TEXT("&v&v"), &vCommand, &vNewStyleFormat);
		if (!Failed(hr, &e, nArgErr))
		{
			_bstr_t bstrVal;
			bstrVal = V_BSTR(&vRet);
			strncpy ((char*)pszResult, (char*)bstrVal, RESULT_LENGTH);
			pszResult[RESULT_LENGTH] = 0;
		}
	}
	catch (CException *e)
	{
		cerr << hex << "Invoke Exception: " << e << endl;
		return E_UNEXPECTED;
	}
	
	return hr;
}



// Untested sample code for creating a new S-PLUS object (e.g., a data.frame or graph)

// BUGBUG: myNewMatrix will dissappear from S-PLUS as soon as it goes out of scope
// (i.e., currently at the end of this routine), so myNewMatrix should actually be created 
// by the parent calling routine, and a pointer to it passed in.
HRESULT AutoClient::CreateMatrix( char* pszNewName, const VARTYPE iMode, LPSAFEARRAY psa)
{
	HRESULT hr;
	EXCEPINFO   e;
	UINT     nArgErr;
	
	AutoObj myNewMatrix;
	COleVariant vSClass( "matrix" );
	COleVariant vRet;
	VariantInit(&vRet);
	
	try
	{		
		hr = SPlusApplication.Invoke(DISPATCH_METHOD, &vRet, &e, &nArgErr,
			OLESTR("CreateObject"), TEXT("&v"), &vSClass);
		if (!Failed(hr, &e, nArgErr))
		{
			
			// Child gets it's own dispatch interface
			myNewMatrix.AttachDispatch( V_DISPATCH(&vRet) );
			
			
			// Find out the new name assigned the matrix by S-PLUS
			VariantInit(&vRet);
			
			hr = myNewMatrix.Invoke(DISPATCH_PROPERTYGET, &vRet, &e, &nArgErr,
				OLESTR("NewName"), NULL, NULL);
			if (!Failed(hr, &e, nArgErr))
			{
				
				USES_CONVERSION;
				strcpy(pszNewName, W2A(V_BSTR(&vRet)));
				
				// Send data to the new object
				COleVariant vData;
				vData.vt = (VT_ARRAY | iMode);
				vData.parray = psa;		
				VariantInit(&vRet);
				
				hr = myNewMatrix.Invoke(DISPATCH_PROPERTYPUT, &vRet, &e, &nArgErr,
					OLESTR("DataAsArray"), TEXT("&v"), &vData);
				Failed(hr, &e, nArgErr);
			}
		}
		
	}
	catch (CException *e)
	{
		cerr << hex << "Invoke Exception: " << e << endl;
		return E_UNEXPECTED;
	}
	
	return hr;
}



inline HRESULT AutoClient::Failed(HRESULT hr, LPEXCEPINFO e, UINT nArgErr)
{
	if SUCCEEDED(hr)
		return FALSE;
	
	if ((hr == DISP_E_PARAMNOTFOUND) || 
		(hr == DISP_E_TYPEMISMATCH))
		cerr << "Trouble occurred with argument number " << nArgErr << endl;
	
	else if (hr == DISP_E_EXCEPTION && e)
	{
		// Server can defer filling this structure
		if (e->pfnDeferredFillIn)
			(*(e->pfnDeferredFillIn))(e);
		
		cerr << hex 
			<< "Invoke Exception: " << e->wCode
			<< "Exception Source: " << e->bstrSource
			<< "Exception Description: " << e->bstrDescription
			<< endl;
	}
	
	else
		cerr << "Unexpected return code " << hex << hr << " received while communicating with S-PLUS." << endl;
	
	// Display message
	#define ERROR_LENGTH  256
	char szError[ERROR_LENGTH+1];
	if (FormatMessage(
        FORMAT_MESSAGE_FROM_SYSTEM,   // | FORMAT_MESSAGE_ALLOCATE_BUFFER
        NULL,
        hr,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        szError,
        ERROR_LENGTH,
        NULL))
		cerr << "System Message: " << szError << " (" << hex << hr << ")." << endl;
	
	return hr;
}



#define PROMPT ">> "
#define CMD_TERM "q()"

UINT AutoClient::ProcessCommands()
{
	HRESULT hr = NULL;
	char pszResult[RESULT_LENGTH+1];
	char command[256];
	UINT nCmds = 0;
	
	// Get and process commands one line at a time 'til user enters CMD_TERM
	cout << "Enter '" << CMD_TERM << "' to terminate this S-PLUS session." << endl;
	
	while(cout << PROMPT, (cin >> command) && stricmp(command, CMD_TERM))
	{	
		hr = ExecuteStringResult((LPCTSTR)command, (LPTSTR)&pszResult);
		
		if FAILED(hr)
		{
			cerr << "ERROR executing '" << command << "'" << endl;
			cerr << hex << "Execute Error: " << hr << endl;
			break;
		}
		
		nCmds++;
		cout << pszResult << endl;
	}
	
	return nCmds;
}



void main()
{
	HRESULT hr;
	AutoClient SClient;
	UINT nCmds;
	
	cout << "Please wait while connection with S-PLUS is established..." << endl;
	
	if FAILED( hr = SClient.CreateSplus() )
		return;
	
	cout << "Connection to S-PLUS established.  Enter each command on one line." << endl;
	
	
	nCmds = SClient.ProcessCommands();
	
	if (nCmds)
		cout << "Processed " << nCmds << " commands this session." << endl;
}